import MemTypes::*;
import EHRPipeReg::*;
import RegFile::*;
import FIFOF::*;

interface DCache;
    method Action req(MemReq r);
    method MemResp resp;
    method Action respDeq;

    method ActionValue#(MemReq) mReq;
    method Action mResp(MemResp r);
endinterface

(* synthesize *)
module mkDCache(DCache);
    RegFile#(Index, LineStatus) sArray <- mkRegFileFull;
    RegFile#(Index, Tag)      tagArray <- mkRegFileFull;
    RegFile#(Index, Data)    dataArray <- mkRegFileFull;

    FIFOF#(MemReq)   mReqQ <- mkFIFOF;
    FIFOF#(MemResp) mRespQ <- mkFIFOF;

    EHRBypassReg#(MemReq) hitQ <- mkEHRBypassReg;
    Reg#(MemReq)       missReq <- mkRegU;
    Reg#(CacheStatus)   status <- mkReg(Rdy);

    rule doStore (hitQ.first.op==St);
        let r = hitQ.first;
        Index idx = truncate(r.addr>>2);
        dataArray.upd(idx, r.data);
        sArray.upd(idx, Dirty);
        hitQ.deq;
    endrule

    rule doWrBack (status==WrBack);
        Index idx = truncate(missReq.addr>>2);
        mReqQ.enq(MemReq{op:St, addr:{tagArray.sub(idx),idx,2'b00}, data:dataArray.sub(idx)});
        status <= FillReq;
    endrule

    rule doFillReq (status==FillReq);
        mReqQ.enq(MemReq{op:Ld, addr:missReq.addr, data:?});
        status <= FillResp;
    endrule

    rule doFillResp (status==FillResp);
        let data = mRespQ.first;  mRespQ.deq;

        Index idx = truncate(missReq.addr>>2);
        Tag tag = truncateLSB(missReq.addr);

        sArray.upd(idx, Clean);
        tagArray.upd(idx, tag);
        dataArray.upd(idx, data);

        status <= FillHit;
    endrule

    rule doFillHit (status==FillHit);
        hitQ.enq(missReq);
        status <= Rdy;
    endrule

    method Action req(MemReq r) if (status==Rdy);
        Index idx = truncate(r.addr>>2);
        Tag tag = truncateLSB(r.addr);
        let lnSt = sArray.sub(idx);
        Bool tagMatch = tagArray.sub(idx)==tag;
        if(lnSt!=Invalid && tagMatch)
            hitQ.enq(r);
        else begin
            missReq <= r;
            status <= lnSt==Dirty ? WrBack : FillReq;
        end
    endmethod

    method MemResp resp if(hitQ.first.op==Ld);
        let r = hitQ.first;
        Index idx = truncate(r.addr>>2);
        return dataArray.sub(idx);
    endmethod

    method Action respDeq if(hitQ.first.op==Ld);
        hitQ.deq;
    endmethod

    method ActionValue#(MemReq) mReq;
        mReqQ.deq;
        return mReqQ.first;
    endmethod

    method Action mResp(MemResp r);
        mRespQ.enq(r);
    endmethod
endmodule
