import ProcTypes::*;
import EHR::*;

interface EHRPipeReg#(type dataT);
    method Action enq(dataT d);
    method Action deq;
    method Action clear;
    method dataT first;
    method Bool notEmpty;
    method Bool notFull;
endinterface

module mkEHRPipeReg(EHRPipeReg#(dataT)) provisos(Bits#(dataT, dataSz));
    Reg#(dataT)   dReg <- mkRegU;
    EHR3#(Bool) vReg <- mkEHR3(False);

    method Action enq(dataT d) if(!vReg.r1);
        dReg <= d;
        vReg.w1(True);
    endmethod
    method Action deq if(vReg.r0);
        vReg.w0(False);
    endmethod
    method Action clear;
        vReg.w2(False);
    endmethod
    method first;
        return dReg;
    endmethod
    method notEmpty = vReg.r0;
    method notFull = !vReg.r1;
endmodule

interface EHRBypassReg#(type dataT);
    method Action enq(dataT d);
    method Action deq;
    method Action clear;
    method dataT first;
    method Bool notEmpty;
    method Bool notFull;
endinterface

module mkEHRBypassReg(EHRBypassReg#(dataT)) provisos(Bits#(dataT, dataSz));
    EHR3#(dataT) dReg <- mkEHR3U;
    EHR3#(Bool)  vReg <- mkEHR3(False);

    method Action enq(dataT d) if(!vReg.r0);
        dReg.w0(d);
        vReg.w0(True);
    endmethod
    method Action deq if(vReg.r1);
        vReg.w1(False);
    endmethod
    method Action clear;
        vReg.w2(False);
    endmethod
    method first;
        return dReg.r1;
    endmethod
    method notEmpty = vReg.r1;
    method notFull = !vReg.r0;
endmodule

interface EHRFIFO#(type t);
    method Action enq(t d);
    method Action deq;
    method t first;
    method Bool notEmpty;
    method Bool notFull;
endinterface

module mkEHRFIFO(EHRFIFO#(t)) provisos(Bits#(t, sz));
  EHR2#(t)    da <- mkEHR2U; 
  EHR2#(Bool) va <- mkEHR2(False);
  EHR2#(t)    db <- mkEHR2U; 
  EHR2#(Bool) vb <- mkEHR2(False);

  rule canonicalize (vb.r1 && !va.r1);
    da.w1(db.r1);
    va.w1(True);
    vb.w1(False);
  endrule          
  method Action enq(t d) if(!vb.r0);
    db.w0(d);
    vb.w0(True);
  endmethod
  method Action deq if(va.r0);
    va.w0(False);
  endmethod
  method t first;
    return da.r0;
  endmethod
  method notEmpty = va.r0;
  method notFull = !vb.r0;
endmodule
