import MemTypes::*;
import EHRPipeReg::*;
import RegFile::*;
import FIFOF::*;

interface ICache;
    method Action req(MemReq r);
    method MemResp resp;
    method Action respDeq;

    method ActionValue#(MemReq) mReq;
    method Action mResp(MemResp r);
endinterface

(* synthesize *)
module mkICache(ICache);
    RegFile#(Index, Bool)    vArray <- mkRegFileFull;
    RegFile#(Index, Tag)   tagArray <- mkRegFileFull;
    RegFile#(Index, Data) dataArray <- mkRegFileFull;

    FIFOF#(MemReq)   mReqQ <- mkFIFOF;
    FIFOF#(MemResp) mRespQ <- mkFIFOF;

    EHRBypassReg#(MemReq) hitQ <- mkEHRBypassReg;
    Reg#(MemReq)       missReq <- mkRegU;
    Reg#(CacheStatus)   status <- mkReg(Rdy);

    rule doFillReq (status==FillReq);
        mReqQ.enq(MemReq{op:Ld, addr:missReq.addr, data:?});
        status <= FillResp;
    endrule

    rule doFillResp (status==FillResp);
        let data = mRespQ.first;  mRespQ.deq;

        Index idx = truncate(missReq.addr>>2);
        Tag tag = truncateLSB(missReq.addr);

        vArray.upd(idx, True);
        tagArray.upd(idx, tag);
        dataArray.upd(idx, data);

        status <= FillHit;
    endrule

    rule doFillHit (status==FillHit);
        hitQ.enq(missReq);
        status <= Rdy;
    endrule

    method Action req(MemReq r) if (status==Rdy);
        Index idx = truncate(r.addr>>2);
        Tag tag = truncateLSB(r.addr);
        Bool valid = vArray.sub(idx);
        Bool tagMatch = tagArray.sub(idx)==tag;
        if(valid && tagMatch)
            hitQ.enq(r);
        else begin
            missReq <= r;
            status <= FillReq;
        end
    endmethod

    method MemResp resp;
        let r = hitQ.first;
        Index idx = truncate(r.addr>>2);
        return dataArray.sub(idx);
    endmethod

    method Action respDeq;
        hitQ.deq;
    endmethod

    method ActionValue#(MemReq) mReq;
        mReqQ.deq;
        return mReqQ.first;
    endmethod

    method Action mResp(MemResp r);
        mRespQ.enq(r);
    endmethod
endmodule
